# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2021 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import aiohttp
import asyncio
import logging
import json
import re
import time
import functools

from typing import TYPE_CHECKING, List, Optional, Any, Union, Tuple
from urllib.parse import quote as urllibquote

from .utils import MaybeLock
from .errors import HTTPException

if TYPE_CHECKING:
    from .client import Client

log = logging.getLogger(__name__)

GRAPHQL_HTML_ERROR_PATTERN = re.compile(
    r'<title>((\d+).*)<\/title>',
    re.MULTILINE
)


def quote(string: str) -> str:
    string = urllibquote(string)
    string = string.replace('/', '%2F')
    return string


class HTTPRetryConfig:
    """Config for how HTTPClient should handle retries.

    .. warning::

        Messing with these values could potentially make retries spammy.
        Worst case scenario of this would be that either your ip or
        your account could be limited due to high traffic. Change these
        values with caution!

    Parameters
    ----------
    max_retry_attempts: :class:`int`
        The max amount of retry attempts for a request. Defaults to ``5``.

        .. note::

            This is ignored when handling capacity throttling.
    max_wait_time: Optional[:class:`float`]
        The max amount of seconds to wait for a request before
        raising the original exception. This works by keeping track of the
        total seconds that has been waited for the request regardless of
        number of attempts. If ``None`` this is ignored. Defaults to ``65``.
    handle_rate_limits: :class:`bool`
        Whether or not the client should handle rate limit errors and wait
        the received ``Retry-After`` before automatically retrying the request.
        Defaults to ``True``.

        .. note::

            This option is only for throttling errors with a Retry-After value.
    max_retry_after: :class:`float`
        The max amount of seconds the client should handle. If a throttled
        error with a higher Retry-After than this value is received, then
        the original :exc:`HTTPException` is raised instead. *Only matters
        when ``handle_rate_limits`` is ``True``*
    other_requests_wait: :class:`bool`
        Whether other requests to a rate limited endpoint should wait
        for the rate limit to disappear before requesting. Defaults to
        ``True``. *Only matters when ``handle_rate_limits`` is ``True``*
    handle_capacity_throttling: :class:`bool`
        Whether or not the client should automatically handle capacity
        throttling errors. These occur when the prod server you
        are requesting from has no available capacity to process a
        request and therefore returns with a throttle error
        without a Retry-After. Defaults to ``True``.
    backoff_start: :class:`float`
        The initial seconds to wait for the exponential backoff. Defaults
        to ``1``. *Only matters when ``handle_capacity_throttling`` is
        ``True``*
    backoff_factor: :class:`float`
        The multiplying factor used for the exponential backoff when a
        request fails. Defaults to ``1.5``. *Only matters when
        ``handle_capacity_throttling`` is ``True``*
    backoff_cap: :class:`float`
        The cap for the exponential backoff to avoid having
        unrealistically high wait times. Defaults to ``20``. *Only matters
        when ``handle_capacity_throttling`` is ``True``*
    """
    def __init__(self, **kwargs):
        self.max_retry_attempts = kwargs.get('max_retry_attempts', 5)
        self.max_wait_time = kwargs.get('max_wait_time', 65)

        self.handle_rate_limits = kwargs.get('handle_rate_limits', True)
        self.max_retry_after = kwargs.get('max_retry_after', 60)
        self.other_requests_wait = kwargs.get('other_requests_wait', True)

        self.handle_capacity_throttling = kwargs.get('handle_capacity_throttling', True)  # noqa
        self.backoff_start = kwargs.get('backoff_start', 1)
        self.backoff_factor = kwargs.get('backoff_factor', 1.5)
        self.backoff_cap = kwargs.get('backoff_cap', 20)


class GraphQLRequest:
    def __init__(self, query: str, *,
                 operation_name: str = None,
                 variables: dict = None
                 ) -> None:
        self.query = query
        self.operation_name = operation_name
        self.variables = variables

    def _to_camel_case(self, text: str) -> str:
        components = text.split('_')
        return components[0] + ''.join(x.title() for x in components[1:])

    def __iter__(self) -> str:
        for key, value in self.__dict__.items():
            if value is None:
                continue

            yield (self._to_camel_case(key), value)

    def as_dict(self) -> dict:
        return dict(self)

    def as_multiple_payload(self) -> dict:
        return {
            'operationName': (self.operation_name
                              or self.get_operation_name_by_query()),
            'variables': self.variables,
            'query': self.query
        }

    def get_operation_name_by_query(self) -> str:
        return re.search(r'(?:mutation|query) (\w+)', self.query).group(1)


class Route:
    """Represents a route to use for a http request. This should
    be subclassed by new routes and the class attributes ``BASE`` and
    optionally ``AUTH`` should be overridden.

    .. warning::

        Usually there is no reason to subclass and implement routes
        yourself as most of them are already implemented. Take a look
        at http.py if you're interested in knowing all of the predefined
        routes.

    Available authentication placeholders:
    - `IOS_BASIC_TOKEN`
    - `FORTNITE_BASIC_TOKEN`
    - `IOS_ACCESS_TOKEN`
    - `FORTNITE_ACCESS_TOKEN`

    Example usage: ::

        class SocialBanPublicService(fortnitepy.Route):
            BASE = 'https://social-ban-public-service-prod.ol.epicgames.com'
            AUTH = 'FORTNITE_ACCESS_TOKEN'

        route = SocialBanPublicService(
            '/socialban/api/public/v1/{user_id}',
            user_id='c7af4984a77a498b83d8b16d475d76bc'
        )
        resp = await client.http.get(route)

        # resp would look something like this:
        # {
        #     "bans" : [],
        #     "warnings" : []
        # }

    Parameters
    ----------
    path: :class:`path`
        The path to used for the request.

        .. warning::

            You should always use name
            formatting for arguments and instead of using `.format()`on the
            path, you should pass the format kwargs as kwargs to the route.
            This might seem counterintuitive but it is important to ensure
            rate limit retry reliability.
    auth: Optional[:class:`str`]
        The authentication to use for the request. If ``None`` the default
        auth specified for the route is used.
    **params: Any
        The variables to format the path with passed alongside their name.

    Attributes
    ----------
    path: :class:`str`
        The requests path.
    params: Dict[:class:`str`, Any]
        A mapping of the params passed.
    base: :class:`str`
        The base of the request url.
    auth: Optional[:class:`str`]
        The auth placeholder.
    url: :class:`str`
        The formatted url.
    sanitized_url: :class:`str`
        The yet to be formatted url.
    """
    BASE = ''
    AUTH = None

    def __init__(self, path: str = '', *,
                 auth: str = None,
                 **params: Any) -> None:
        self.path = path
        self.params = {k: (quote(v) if isinstance(v, str) else v)
                       for k, v in params.items()}

        if self.BASE == '':
            raise ValueError('Route must have a base')

        self.sanitized_url = url = self.BASE + self.path
        self.url = url.format(**self.params) if self.params else url

        if auth:
            self.AUTH = auth

        self.base = self.BASE
        self.auth = self.AUTH


class EpicGamesGraphQL(Route):
    BASE = 'https://graphql.epicgames.com/graphql'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class EpicGames(Route):
    BASE = 'https://www.epicgames.com'
    AUTH = None


class PaymentWebsite(Route):
    BASE = 'https://payment-website-pci.ol.epicgames.com'
    AUTH = None


class LightswitchPublicService(Route):
    BASE = 'https://lightswitch-public-service-prod06.ol.epicgames.com'
    AUTH = 'IOS_ACCESS_TOKEN'


class UserSearchService(Route):
    BASE = 'https://user-search-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class AccountPublicService(Route):
    BASE = 'https://account-public-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class EulatrackingPublicService(Route):
    BASE = 'https://eulatracking-public-service-prod-m.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class AffiliatePublicService(Route):
    BASE = 'https://affiliate-public-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class EventsPublicService(Route):
    BASE = 'https://events-public-service-live.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class FortniteContentWebsite(Route):
    BASE = 'https://fortnitecontent-website-prod07.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class FortnitePublicService(Route):
    BASE = 'https://fortnite-public-service-prod11.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class FriendsPublicService(Route):
    BASE = 'https://friends-public-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class PartyService(Route):
    BASE = 'https://party-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class PresencePublicService(Route):
    BASE = 'https://presence-public-service-prod.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


class StatsproxyPublicService(Route):
    BASE = 'https://statsproxy-public-service-live.ol.epicgames.com'
    AUTH = 'FORTNITE_ACCESS_TOKEN'


def create_aiohttp_closed_event(session) -> asyncio.Event:
    """Work around aiohttp issue that doesn't properly close transports on exit.

    See https://github.com/aio-libs/aiohttp/issues/1925#issuecomment-639080209

    Returns:
       An event that will be set once all transports have been properly closed.
    """

    transports = 0
    all_is_lost = asyncio.Event()

    def connection_lost(exc, orig_lost):
        nonlocal transports

        try:
            orig_lost(exc)
        finally:
            transports -= 1
            if transports == 0:
                all_is_lost.set()

    def eof_received(orig_eof_received):
        try:
            orig_eof_received()
        except AttributeError:
            # It may happen that eof_received() is called after
            # _app_protocol and _transport are set to None.
            pass

    for conn in session.connector._conns.values():
        for handler, _ in conn:
            proto = getattr(handler.transport, "_ssl_protocol", None)
            if proto is None:
                continue

            transports += 1
            orig_lost = proto.connection_lost
            orig_eof_received = proto.eof_received

            proto.connection_lost = functools.partial(
                connection_lost, orig_lost=orig_lost
            )
            proto.eof_received = functools.partial(
                eof_received, orig_eof_received=orig_eof_received
            )

    if transports == 0:
        all_is_lost.set()

    return all_is_lost


class HTTPClient:
    def __init__(self, client: 'Client', *,
                 connector: aiohttp.BaseConnector = None,
                 retry_config: Optional[HTTPRetryConfig] = None) -> None:
        self.client = client
        self.connector = connector
        self.retry_config = retry_config or HTTPRetryConfig()

        self._jar = aiohttp.CookieJar()
        self.headers = {}
        self.device_id = self.client.auth.device_id
        self._endpoint_events = {}

        # How many refreshes (max_refresh_attempts) to attempt in
        # a time window (refresh_attempt_window) before closing.
        self.max_refresh_attempts = 3
        self.refresh_attempt_window = 20

        self.create_connection()

    @staticmethod
    async def json_or_text(response: aiohttp.ClientResponse) -> Union[str,
                                                                      dict]:
        text = await response.text(encoding='utf-8')
        if 'application/json' in response.headers.get('content-type', ''):
            return json.loads(text)
        return text

    @property
    def user_agent(self) -> str:
        return 'Fortnite/{0.client.build} {0.client.os}'.format(self)

    def get_auth(self, auth: str) -> str:
        u_auth = auth.upper()

        if u_auth == 'IOS_BASIC_TOKEN':
            return 'basic {0}'.format(self.client.auth.ios_token)
        elif u_auth == 'FORTNITE_BASIC_TOKEN':
            return 'basic {0}'.format(self.client.auth.fortnite_token)
        elif u_auth == 'IOS_ACCESS_TOKEN':
            return self.client.auth.ios_authorization
        elif u_auth == 'FORTNITE_ACCESS_TOKEN':
            return self.client.auth.authorization
        return auth

    def add_header(self, key: str, val: Any) -> None:
        self.headers[key] = val

    def remove_header(self, key: str) -> Any:
        return self.headers.pop(key)

    async def close(self) -> None:
        self._jar.clear()
        if self.__session:
            event = create_aiohttp_closed_event(self.__session)
            await self.__session.close()
            try:
                await asyncio.wait_for(event.wait(), timeout=2)
            except asyncio.TimeoutError:
                pass

    def create_connection(self) -> None:
        self.__session = aiohttp.ClientSession(
            connector=self.connector,
            connector_owner=self.connector is None,
            cookie_jar=self._jar
        )

    async def request(self, method: str, url: str,
                      **kwargs: Any
                      ) -> Tuple[aiohttp.ClientResponse, Union[str, dict]]:
        try:
            params = kwargs['params']
            if isinstance(params, dict):
                kwargs['params'] = {k: (str(v).lower() if isinstance(v, bool)
                                        else v) for k, v in params.items()}
            else:
                kwargs['params'] = [(k, (str(v).lower() if isinstance(v, bool)
                                         else v)) for k, v in params]
        except KeyError:
            pass

        pre_time = time.time()
        async with self.__session.request(method, url, **kwargs) as r:
            log.debug('{0} {1} has returned {2.status} in {3:.2f}s'.format(
                method,
                url,
                r,
                time.time() - pre_time
            ))

            data = await self.json_or_text(r)
            return r, data

    async def _fn_request(self, method: str,
                          route: Union[Route, str],
                          auth: Optional[str] = None,
                          graphql: Optional[Union[Route, List[Route]]] = None,
                          **kwargs: Any) -> Any:
        url = route.url if not isinstance(route, str) else route

        headers = {**kwargs.get('headers', {}), **self.headers}
        headers['User-Agent'] = self.user_agent

        auth = auth or route.AUTH
        if auth is not None:
            headers['Authorization'] = self.get_auth(auth)

        device_id = kwargs.pop('device_id', None)
        if device_id is not None:
            headers['X-Epic-Device-ID'] = (self.device_id if device_id is True
                                           else device_id)

        if graphql is not None:
            is_multiple = isinstance(graphql, (list, tuple))

            if not is_multiple:
                graphql = (graphql,)
            kwargs['json'] = [gql_query.as_multiple_payload()
                              for gql_query in graphql]

        kwargs['headers'] = headers

        try:
            del kwargs['config']
        except KeyError:
            pass

        raw = kwargs.pop('raw', False)
        r, data = await self.request(method, url, **kwargs)

        if raw:
            return r

        if graphql is not None:
            if isinstance(data, str):
                m = GRAPHQL_HTML_ERROR_PATTERN.search(data)
                error_data = ({
                    'serviceResponse': '',
                    'message': 'Unknown reason' if m is None else m.group(1)
                },)
                if m is not None:
                    error_data[0]['serviceResponse'] = json.dumps({
                        'errorStatus': int(m.group(2))
                    })

            elif isinstance(data, dict):
                if data['status'] >= 400:
                    message = data['message']
                    error_data = ({
                        'serviceResponse': json.dumps({
                            'errorCode': message
                        }),
                        'message': message
                    },)
            else:
                error_data = None
                for child_data in data:
                    if 'errors' in child_data:
                        error_data = child_data['errors']
                        break

            if error_data is not None:
                selected = error_data[0]

                obj = {'errorMessage': selected['message']}
                service_response = selected['serviceResponse']
                if service_response == '':
                    error_payload = {}
                else:
                    error_payload = json.loads(service_response)

                if isinstance(error_payload, str):
                    m = GRAPHQL_HTML_ERROR_PATTERN.search(error_payload)
                    message = 'Unknown reason' if m is None else m.group(1)
                    error_payload = {
                        'errorMessage': message,
                    }

                    if m is not None:
                        error_payload['errorStatus'] = int(m.group(2))

                raise HTTPException(
                    r,
                    route,
                    {**obj, **error_payload},
                    headers
                )

            def get_payload(d):
                return next(iter(d['data'].values()))

            if len(data) == 1:
                return get_payload(data[0])
            return [get_payload(d) for d in data]

        if 'errorCode' in data or r.status >= 400:
            if isinstance(data, str):
                data = {
                    'errorMessage': data if data else 'Unknown {}'.format(
                        r.status
                    )
                }
            raise HTTPException(r, route, data, headers)

        return data

    def get_retry_after(self, exc):
        retry_after = exc.response.headers.get('Retry-After')
        if retry_after is not None:
            return int(retry_after)

        # For some reason graphql response headers doesn't contain
        # rate limit headers so we have to get retry after from
        # the message vars.
        try:
            return int(exc.message_vars[0])
        except (ValueError, IndexError):
            return None

    async def fn_request(self, method: str,
                         route: Union[Route, str],
                         auth: Optional[str] = None,
                         graphql: Union[Route, List[Route]] = None,
                         priority: int = 0,
                         **kwargs: Any) -> Any:
        if self.client.is_closed():
            raise RuntimeError('Client is closed.')

        cfg = self.retry_config
        if isinstance(route, Route):
            url = route.url
            url_key = (method, route.sanitized_url)
        else:
            url = route
            url_key = None

        tries = 0
        total_slept = 0
        backoff = cfg.backoff_start
        while True:
            sleep_time = 0
            tries += 1

            endpoint_event = self._endpoint_events.get(url_key)
            if endpoint_event is not None:
                log.debug('Waiting for {0:.2f}s before requesting {1} {2}.'.format(  # noqa
                    endpoint_event.ends_at - time.time(),
                    method,
                    url,
                ))
                await endpoint_event.wait()

            endpoint_event = None

            lock = self.client._reauth_lock
            if priority <= 0:
                await lock.wait()
                if lock.failed:
                    raise asyncio.CancelledError(
                        'Client is shutting down.'
                    )

            try:
                return await self._fn_request(
                    method,
                    route,
                    auth,
                    graphql,
                    **kwargs
                )
            except HTTPException as exc:
                if self.client._closing:
                    raise

                if tries >= cfg.max_retry_attempts:
                    raise

                code = exc.message_code

                if graphql:
                    gql_server_error = exc.raw.get('errorStatus') in {500, 502}
                else:
                    gql_server_error = False

                catch = (
                    'errors.com.epicgames.common.oauth.invalid_token',
                    'errors.com.epicgames.common.authentication.token_verification_failed',  # noqa
                    'error.graphql.401',
                )
                if code in catch:
                    _auth = auth or route.AUTH
                    if exc.request_headers['Authorization'] != self.get_auth(_auth):  # noqa
                        continue

                    force_attempts = self.max_refresh_attempts
                    retry = True

                    def should_force():
                        ts = self.client._refresh_times
                        if len(ts) > force_attempts:
                            self.client._refresh_times = ts[-force_attempts:]
                        try:
                            old = self.client._refresh_times[-force_attempts]
                        except IndexError:
                            return True
                        else:
                            cur = time.time()
                            return cur - old > self.refresh_attempt_window

                    if priority > lock.priority - 1:
                        async with MaybeLock(lock):
                            if should_force():
                                try:
                                    await self.client.auth.do_refresh()
                                except asyncio.CancelledError:
                                    lock.failed = True
                                    retry = False
                                except Exception:
                                    if self.client.can_restart():
                                        await self.client.restart()
                                    else:
                                        lock.failed = True
                                        retry = False
                            else:
                                retry = False
                    else:
                        if lock.locked():
                            await lock.wait()
                            if lock.failed:
                                raise asyncio.CancelledError(
                                    'Client is shutting down.'
                                )
                        else:
                            retry = False

                    if retry:
                        continue
                    else:
                        try:
                            e = RuntimeError('Oauth token invalid.')
                            e.__cause__ = exc
                            self.client._exception_future.set_exception(e)
                        except asyncio.InvalidStateError:
                            pass

                        raise asyncio.CancelledError(
                            'Client is shutting down.'
                        )

                elif code == 'errors.com.epicgames.common.throttled' or exc.status == 429:  # noqa
                    retry_after = self.get_retry_after(exc)
                    if retry_after is not None and cfg.handle_rate_limits:
                        if retry_after <= cfg.max_retry_after:
                            sleep_time = retry_after + 0.5
                            if cfg.other_requests_wait and url_key is not None:
                                if url_key not in self._endpoint_events:
                                    endpoint_event = asyncio.Event()
                                    endpoint_event.ends_at = time.time() + sleep_time  # noqa
                                    self._endpoint_events[url_key] = endpoint_event  # noqa
                    else:
                        tries -= 1  # backoff tries shouldn't count
                        if cfg.handle_capacity_throttling:
                            backoff *= cfg.backoff_factor
                            if backoff <= cfg.backoff_cap:
                                sleep_time = backoff

                elif (code == 'errors.com.epicgames.common.concurrent_modification_error'  # noqa
                        or code == 'errors.com.epicgames.common.server_error'
                        or gql_server_error):  # noqa
                    sleep_time = 0.5 + (tries - 1) * 2

                if sleep_time > 0:
                    total_slept += sleep_time
                    if cfg.max_wait_time and total_slept > cfg.max_wait_time:
                        raise

                    log.debug('Retrying {0} {1} in {2:.2f}s.'.format(
                        method,
                        url,
                        sleep_time
                    ))
                    await asyncio.sleep(sleep_time)
                    continue
                raise

            except aiohttp.ServerDisconnectedError:
                await asyncio.sleep(0.5 + (tries - 1) * 2)
                continue
            except OSError as exc:
                if exc.errno in (54, 10054):
                    continue
                raise

            finally:
                if endpoint_event is not None:
                    del self._endpoint_events[url_key]
                    endpoint_event.set()

    async def get(self, route: Union[Route, str],
                  auth: Optional[str] = None,
                  **kwargs: Any) -> Any:
        return await self.fn_request('GET', route, auth, **kwargs)

    async def post(self, route: Union[Route, str],
                   auth: Optional[str] = None,
                   **kwargs: Any) -> Any:
        return await self.fn_request('POST', route, auth, **kwargs)

    async def delete(self, route: Union[Route, str],
                     auth: Optional[str] = None,
                     **kwargs: Any) -> Any:
        return await self.fn_request('DELETE', route, auth, **kwargs)

    async def patch(self, route: Union[Route, str],
                    auth: Optional[str] = None,
                    **kwargs: Any) -> Any:
        return await self.fn_request('PATCH', route, auth, **kwargs)

    async def put(self, route: Union[Route, str],
                  auth: Optional[str] = None,
                  **kwargs: Any) -> Any:
        return await self.fn_request('PUT', route, auth, **kwargs)

    async def graphql_request(self, graphql: Union[GraphQLRequest,
                                                   List[GraphQLRequest]],
                              auth: Optional[str] = None,
                              **kwargs: Any) -> Any:
        return await self.fn_request('POST', EpicGamesGraphQL(), auth, graphql,
                                     **kwargs)

    ###################################
    #        Epicgames GraphQL        #
    ###################################

    async def graphql_friends_set_alias(self) -> Any:
        variables = {
            "friendId": "65db72079052463cb345d23ee27ae6a1",
            "alias": "Hallo1233"
        }

        query = """
        mutation FriendsMutation($friendId: String!, $alias: String!) {
            Friends {
                setAlias(friendId: $friendId, alias: $alias) {
                    success
                }
            }
        }"""

        return await self.graphql_request(
            GraphQLRequest(query, variables=variables))

    async def graphql_initialize_friends_request(self) -> List[dict]:
        queries = (
            GraphQLRequest(
                query="""
                query FriendsQuery($displayNames: Boolean!) {
                    Friends {
                        summary(displayNames: $displayNames) {
                            friends {
                                alias
                                note
                                favorite
                                ...friendFields
                            }
                            incoming {
                                ...friendFields
                            }
                            outgoing {
                                ...friendFields
                            }
                            blocklist {
                                ...friendFields
                            }
                        }
                    }
                }

                fragment friendFields on Friend {
                    accountId
                    displayName
                    account {
                        externalAuths {
                            type
                            accountId
                            externalAuthId
                            externalDisplayName
                        }
                    }
                }
                """,
                variables={
                    'displayNames': True
                }
            ),
            GraphQLRequest(
                query="""
                query PresenceV2Query($namespace: String!, $circle: String!) {
                    PresenceV2 {
                        getLastOnlineSummary(namespace: $namespace, circle: $circle) {
                            summary { #Type: [LastOnline]
                                friendId #Type: String
                                last_online #Type: String
                            }
                        }
                    }
                }
                """,  # noqa
                variables={
                    'namespace': 'Fortnite',
                    'circle': 'friends'
                }
            )
        )

        return await self.graphql_request(queries)

    ###################################
    #            Epicgames            #
    ###################################

    async def epicgames_get_csrf(self) -> aiohttp.ClientResponse:
        return await self.get(EpicGames('/id/api/csrf'), raw=True)

    async def epicgames_reputation(self, xsrf_token: str) -> Any:
        headers = {
            'x-xsrf-token': xsrf_token
        }

        return await self.get(EpicGames('/id/api/reputation'), headers=headers)

    async def epicgames_login(self, email: str,
                              password: str,
                              xsrf_token: str) -> Any:
        headers = {
            'x-xsrf-token': xsrf_token
        }

        payload = {
            'email': email,
            'password': password,
            'rememberMe': False,
            'captcha': ''
        }

        return await self.post(EpicGames('/id/api/login'),
                               headers=headers,
                               data=payload)

    async def epicgames_mfa_login(self, method: str,
                                  code: str,
                                  xsrf_token: str) -> Any:
        headers = {
            'x-xsrf-token': xsrf_token
        }

        payload = {
            'code': code,
            'method': method,
            'rememberDevice': False
        }

        return await self.post(EpicGames('/id/api/login/mfa'),
                               headers=headers,
                               data=payload)

    async def epicgames_redirect(self, xsrf_token: str) -> Any:
        headers = {
            'x-xsrf-token': xsrf_token
        }

        return await self.get(EpicGames('/id/api/redirect'), headers=headers)

    async def epicgames_get_exchange_data(self, xsrf_token: str) -> dict:
        headers = {
            'x-xsrf-token': xsrf_token
        }

        r = EpicGames('/id/api/exchange/generate')
        return await self.post(r, headers=headers)

    ###################################
    #        Payment Website          #
    ###################################

    async def payment_website_search_sac_by_slug(self, slug: str) -> Any:
        params = {
            'slug': slug
        }

        r = PaymentWebsite('/affiliate/search-by-slug', auth=None)
        return await self.get(r, params=params)

    ###################################
    #           Lightswitch           #
    ###################################

    async def lightswitch_get_status(self, *,
                                     service_id: Optional[str] = None) -> list:
        params = {'serviceId': service_id} if service_id else None

        r = LightswitchPublicService('/lightswitch/api/service/bulk/status')
        return await self.get(r, params=params)

    ###################################
    #           User Search           #
    ###################################

    async def user_search_by_prefix(self, prefix: str, platform: str) -> list:
        params = {
            'prefix': prefix,
            'platform': platform
        }

        r = UserSearchService('/api/v1/search')
        return await self.get(r, params=params)

    ###################################
    #            Account              #
    ###################################

    async def account_get_exchange_data(self, auth: str,
                                        **kwargs: Any) -> dict:
        r = AccountPublicService('/account/api/oauth/exchange')
        return await self.get(r, auth=auth, **kwargs)

    async def account_oauth_grant(self, **kwargs: Any) -> dict:
        r = AccountPublicService('/account/api/oauth/token')
        return await self.post(r, **kwargs)

    async def account_generate_device_auth(self, client_id: str) -> dict:
        r = AccountPublicService(
            '/account/api/public/account/{client_id}/deviceAuth',
            client_id=client_id
        )
        return await self.post(r, auth="IOS_ACCESS_TOKEN", json={})

    async def account_get_device_auths(self, client_id: str) -> list:
        r = AccountPublicService(
            '/account/api/public/account/{client_id}/deviceAuth',
            client_id=client_id,
            auth="IOS_ACCESS_TOKEN"
        )
        return await self.get(r)

    async def account_lookup_device_auth(self, client_id: str,
                                         device_id: str) -> dict:
        r = AccountPublicService(
            '/account/api/public/account/{client_id}/deviceAuth/{device_id}',
            client_id=client_id,
            device_id=device_id,
            auth="IOS_ACCESS_TOKEN"
        )
        return await self.get(r)

    async def account_delete_device_auth(self, client_id: str,
                                         device_id: str) -> None:
        r = AccountPublicService(
            '/account/api/public/account/{client_id}/deviceAuth/{device_id}',
            client_id=client_id,
            device_id=device_id,
            auth="IOS_ACCESS_TOKEN"
        )
        return await self.delete(r)

    async def account_sessions_kill_token(self, token: str, auth=None) -> Any:
        r = AccountPublicService(
            '/account/api/oauth/sessions/kill/{token}',
            token=token
        )
        return await self.delete(r, auth='bearer {0}'.format(token))

    async def account_sessions_kill(self, kill_type: str,
                                    auth='IOS_ACCESS_TOKEN',
                                    **kwargs: Any) -> Any:
        params = {
            'killType': kill_type
        }

        r = AccountPublicService('/account/api/oauth/sessions/kill')
        return await self.delete(r, params=params, auth=auth, **kwargs)

    async def account_get_by_display_name(self, display_name: str) -> dict:
        r = AccountPublicService(
            '/account/api/public/account/displayName/{display_name}',
            display_name=display_name
        )
        return await self.get(r)

    async def account_get_by_user_id(self, user_id: str, *,
                                     auth: Optional[str] = None,
                                     **kwargs: Any) -> dict:
        r = AccountPublicService(
            '/account/api/public/account/{user_id}',
            user_id=user_id
        )
        return await self.get(r, auth=auth, **kwargs)

    async def account_get_by_email(self, email: str) -> dict:
        r = AccountPublicService(
            '/account/api/public/account/email/{email}',
            email=email
        )
        return await self.get(r, auth='IOS_ACCESS_TOKEN')

    async def account_get_external_auths_by_id(self, user_id: str,
                                               **kwargs: Any) -> list:
        r = AccountPublicService(
            '/account/api/public/account/{user_id}/externalAuths',
            user_id=user_id
        )
        return await self.get(r, **kwargs)

    async def account_get_multiple_by_user_id(self,
                                              user_ids: List[str]) -> list:
        params = [('accountId', user_id) for user_id in user_ids]
        r = AccountPublicService('/account/api/public/account')
        return await self.get(r, params=params)

    async def account_graphql_get_multiple_by_user_id(self,
                                                      user_ids: List[str],
                                                      **kwargs: Any
                                                      ) -> dict:
        return await self.graphql_request(GraphQLRequest(
            query="""
            query AccountQuery($accountIds: [String]!) {
                Account {
                    # Get details about an account given an account ID
                    accounts(accountIds: $accountIds) { #Type: [Account]
                        # The AccountID for this account
                        id #Type: String
                        # The epic display name for this account
                        displayName #Type: String
                        # External auths associated with this account
                        externalAuths { #Type: [ExternalAuth]
                            type #Type: String
                            accountId #Type: String
                            externalAuthId #Type: String
                            externalDisplayName #Type: String
                        }
                    }
                }
            }
            """,
            variables={
                'accountIds': user_ids
            }
        ), **kwargs)

    async def account_graphql_get_by_display_name(self,
                                                  display_name: str) -> dict:
        return await self.graphql_request(GraphQLRequest(
            query="""
            query AccountQuery($displayName: String!) {
                Account {
                    account(displayName: $displayName) {
                        id
                        displayName
                        externalAuths {
                            type
                            accountId
                            externalAuthId
                            externalDisplayName
                        }
                    }
                }
            }
            """,
            variables={
                'displayName': display_name
            }
        ))

    async def account_graphql_get_clients_external_auths(self,
                                                         **kwargs: Any
                                                         ) -> dict:
        return await self.graphql_request(GraphQLRequest(
            query="""
            query AccountQuery {
                Account {
                    myAccount {
                        externalAuths {
                            type
                            accountId
                            externalAuthId
                            externalDisplayName
                        }
                    }
                }
            }
            """
        ), **kwargs)

    ###################################
    #          Eula Tracking          #
    ###################################

    async def eulatracking_get_data(self, **kwargs: Any) -> dict:
        r = EulatrackingPublicService(
            '/eulatracking/api/public/agreements/fn/account/{client_id}',
            client_id=self.client.user.id
        )
        return await self.get(r, **kwargs)

    async def eulatracking_accept(self, version: int, *,
                                  locale: str = 'en',
                                  **kwargs: Any) -> Any:
        params = {
            'locale': locale
        }

        r = EulatrackingPublicService(
            ('/eulatracking/api/public/agreements/fn/version/{version}'
             '/account/{client_id}/accept'),
            version=version,
            client_id=self.client.user.id
        )
        return await self.post(r, params=params, **kwargs)

    ###################################
    #         Fortnite Public         #
    ###################################

    async def fortnite_grant_access(self, **kwargs: Any) -> Any:
        r = FortnitePublicService(
            '/fortnite/api/game/v2/grant_access/{client_id}',
            client_id=self.client.user.id
        )
        return await self.post(r, json={}, **kwargs)

    async def fortnite_get_store_catalog(self) -> dict:
        r = FortnitePublicService('/fortnite/api/storefront/v2/catalog')
        return await self.get(r)

    async def fortnite_check_gift_eligibility(self,
                                              user_id: str,
                                              offer_id: str) -> Any:
        r = FortnitePublicService(
            '/fortnite/api/storefront/v2/gift/check_eligibility/recipient/{user_id}/offer/{offer_id}',  # noqa
            user_id=user_id,
            offer_id=offer_id,
        )
        return await self.get(r)

    async def fortnite_get_timeline(self) -> dict:
        r = FortnitePublicService('/fortnite/api/calendar/v1/timeline')
        return await self.get(r)

    ###################################
    #        Fortnite Content         #
    ###################################

    async def fortnitecontent_get(self) -> dict:
        r = FortniteContentWebsite('/content/api/pages/fortnite-game')
        return await self.get(r)

    ###################################
    #            Friends              #
    ###################################

    async def friends_get_all(self, *,
                              include_pending: bool = False,
                              **kwargs) -> list:
        params = {
            'includePending': include_pending
        }

        r = FriendsPublicService('/friends/api/public/friends/{client_id}',
                                 client_id=self.client.user.id)
        return await self.get(r, params=params, **kwargs)

    async def friends_add_or_accept(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.post(r)

    async def friends_remove_or_decline(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.delete(r)

    async def friends_get_blocklist(self) -> list:
        r = FriendsPublicService('/friends/api/v1/{client_id}/blocklist',
                                 client_id=self.client.user.id)
        return await self.get(r)

    async def friends_set_nickname(self, user_id: str, nickname: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}/alias',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.put(r, data=nickname)

    async def friends_remove_nickname(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}/alias',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.delete(r)

    async def friends_set_note(self, user_id: str, note: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}/note',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.put(r, data=note)

    async def friends_remove_note(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}/note',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.delete(r)

    async def friends_get_summary(self, **kwargs) -> dict:
        r = FriendsPublicService('/friends/api/v1/{client_id}/summary',
                                 client_id=self.client.user.id)
        return await self.get(r, **kwargs)

    async def friends_block(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/blocklist/{user_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.post(r)

    async def friends_unblock(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/blocklist/{user_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.delete(r)

    async def friends_get_mutual(self, user_id: str) -> Any:
        r = FriendsPublicService(
            '/friends/api/v1/{client_id}/friends/{user_id}/mutual',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.get(r)

    ###################################
    #            Presence             #
    ###################################

    async def presence_get_last_online(self, **kwargs) -> dict:
        r = PresencePublicService('/presence/api/v1/_/{client_id}/last-online',
                                  client_id=self.client.user.id)
        return await self.get(r, **kwargs)

    ###################################
    #              Stats              #
    ###################################

    async def stats_get_v2(self, user_id: str, *,
                           start_time: Optional[int] = None,
                           end_time: Optional[int] = None) -> dict:
        params = {}
        if start_time:
            params['startTime'] = start_time
        if end_time:
            params['endTime'] = end_time

        r = StatsproxyPublicService(
            '/statsproxy/api/statsv2/account/{user_id}',
            user_id=user_id
        )
        return await self.get(r, params=params)

    async def stats_get_multiple_v2(self, ids: List[str], stats: List[str], *,
                                    category: Optional[str] = None,
                                    start_time: Optional[int] = None,
                                    end_time: Optional[int] = None) -> list:
        payload = {
            'appId': 'fortnite',
            'owners': ids,
            'stats': stats
        }

        params = {}
        if category is not None:
            params['category'] = 'collection_fish'

        if start_time:
            payload['startDate'] = start_time
        if end_time:
            payload['endDate'] = end_time

        r = StatsproxyPublicService('/statsproxy/api/statsv2/query')
        return await self.post(r, json=payload, params=params)

    async def stats_get_leaderboard_v2(self, stat: str) -> dict:
        r = StatsproxyPublicService(
            '/statsproxy/api/statsv2/leaderboards/{stat}',
            stat=stat
        )
        return await self.get(r)

    ###################################
    #             Party               #
    ###################################

    async def party_disconnect(self, party_id: str, user_id: str):
        r = PartyService(
            'party/api/v1/Fortnite/parties/{party_id}/members/{user_id}/disconnect',  # noqa
            party_id=party_id,
            user_id=user_id,
        )
        return await self.post(r)

    async def party_send_invite(self, party_id: str,
                                user_id: str,
                                send_ping: bool = True) -> Any:
        conn_type = self.client.default_party_member_config.cls.CONN_TYPE
        payload = {
            'urn:epic:cfg:build-id_s': self.client.party_build_id,
            'urn:epic:conn:platform_s': self.client.platform.value,
            'urn:epic:conn:type_s': conn_type,
            'urn:epic:invite:platformdata_s': '',
            'urn:epic:member:dn_s': self.client.user.display_name,
        }

        params = {
            'sendPing': send_ping
        }

        r = PartyService(
            '/party/api/v1/Fortnite/parties/{party_id}/invites/{user_id}',
            party_id=party_id,
            user_id=user_id
        )
        return await self.post(r, json=payload, params=params)

    async def party_delete_invite(self, party_id: str, user_id: str) -> Any:
        r = PartyService(
            '/party/api/v1/Fortnite/parties/{party_id}/invites/{user_id}',
            party_id=party_id,
            user_id=user_id
        )
        return await self.delete(r)

    # NOTE: Depracated since fortnite v11.30. Use param sendPing=True with
    #       send_invite
    # NOTE: Now used for sending invites from private parties
    async def party_send_ping(self, user_id: str) -> Any:
        r = PartyService(
            '/party/api/v1/Fortnite/user/{user_id}/pings/{client_id}',
            user_id=user_id,
            client_id=self.client.user.id
        )
        return await self.post(r, json={})

    async def party_delete_ping(self, user_id: str) -> Any:
        r = PartyService(
            '/party/api/v1/Fortnite/user/{client_id}/pings/{user_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.delete(r)

    async def party_decline_invite(self, party_id: str) -> Any:
        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/invites/'
             '{client_id}/decline'),
            party_id=party_id,
            client_id=self.client.user.id
        )
        return await self.post(r, json={})

    async def party_member_confirm(self, party_id: str, user_id: str) -> Any:
        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/members/'
             '{user_id}/confirm'),
            party_id=party_id,
            user_id=user_id
        )
        return await self.post(r, json={})

    async def party_member_reject(self, party_id: str, user_id: str) -> Any:
        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/members/'
             '{user_id}/reject'),
            party_id=party_id,
            user_id=user_id
        )
        return await self.post(r, json={})

    async def party_promote_member(self, party_id: str, user_id: str) -> Any:
        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/members/'
             '{user_id}/promote'),
            party_id=party_id,
            user_id=user_id
        )
        return await self.post(r, json={})

    async def party_kick_member(self, party_id: str, user_id: str) -> Any:
        r = PartyService(
            '/party/api/v1/Fortnite/parties/{party_id}/members/{user_id}',
            party_id=party_id,
            user_id=user_id
        )
        return await self.delete(r)

    async def party_leave(self, party_id: str, **kwargs: Any) -> Any:
        conn_type = self.client.default_party_member_config.cls.CONN_TYPE
        payload = {
            'connection': {
                'id': self.client.user.jid,
                'meta': {
                    'urn:epic:conn:platform_s': self.client.platform.value,
                    'urn:epic:conn:type_s': conn_type,
                },
            },
            'meta': {
                'urn:epic:member:dn_s': self.client.user.display_name,
                'urn:epic:member:type_s': conn_type,
                'urn:epic:member:platform_s': self.client.platform.value,
                'urn:epic:member:joinrequest_j': json.dumps({
                    'CrossplayPreference_i': '1'
                }),
            }
        }

        r = PartyService(
            '/party/api/v1/Fortnite/parties/{party_id}/members/{client_id}',
            party_id=party_id,
            client_id=self.client.user.id
        )
        return await self.delete(r, json=payload, **kwargs)

    async def party_join_request(self, party_id: str) -> Any:
        conf = self.client.default_party_member_config
        conn_type = conf.cls.CONN_TYPE
        payload = {
            'connection': {
                'id': str(self.client.xmpp.xmpp_client.local_jid),
                'meta': {
                    'urn:epic:conn:platform_s': self.client.platform.value,
                    'urn:epic:conn:type_s': conn_type,
                },
                'yield_leadership': conf.yield_leadership,
                'offline_ttl': conf.offline_ttl,
            },
            'meta': {
                'urn:epic:member:dn_s': self.client.user.display_name,
                'urn:epic:member:joinrequestusers_j': json.dumps({
                    'users': [
                        {
                            'id': self.client.user.id,
                            'dn': self.client.user.display_name,
                            'plat': self.client.platform.value,
                            'data': json.dumps({
                                'CrossplayPreference': '1',
                                'SubGame_u': '1',
                            })
                        }
                    ]
                }),
            },
        }

        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/members/'
             '{client_id}/join'),
            party_id=party_id,
            client_id=self.client.user.id
        )
        return await self.post(r, json=payload)

    async def party_send_intention(self, user_id: str) -> dict:
        payload = {
            'urn:epic:invite:platformdata_s': '',
        }

        r = PartyService(
            '/party/api/v1/Fortnite/members/{user_id}/intentions/{client_id}',
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.post(r, json=payload)

    async def party_lookup(self, party_id: str, **kwargs: Any) -> dict:
        r = PartyService('/party/api/v1/Fortnite/parties/{party_id}',
                         party_id=party_id)
        return await self.get(r, **kwargs)

    async def party_lookup_user(self, user_id: str, **kwargs: Any) -> dict:
        r = PartyService('/party/api/v1/Fortnite/user/{user_id}',
                         user_id=user_id)
        return await self.get(r, **kwargs)

    async def party_lookup_ping(self, user_id: str) -> list:
        r = PartyService(
            ('/party/api/v1/Fortnite/user/{client_id}/pings/'
             '{user_id}/parties'),
            client_id=self.client.user.id,
            user_id=user_id
        )
        return await self.get(r)

    async def party_create(self, config: dict, **kwargs: Any) -> dict:
        conf = self.client.default_party_member_config
        conn_type = conf.cls.CONN_TYPE

        _chat_enabled = str(config['chat_enabled']).lower()
        payload = {
            'config': {
                'join_confirmation': config['join_confirmation'],
                'joinability': config['joinability'],
                'max_size': config['max_size']
            },
            'join_info': {
                'connection': {
                    'id': str(self.client.xmpp.xmpp_client.local_jid),
                    'meta': {
                        'urn:epic:conn:platform_s': self.client.platform.value,
                        'urn:epic:conn:type_s': conn_type
                    },
                    'yield_leadership': conf.yield_leadership,
                    'offline_ttl': conf.offline_ttl,
                },
            },
            'meta': {
                'urn:epic:cfg:accepting-members_b': False,
                'urn:epic:cfg:build-id_s': str(self.client.party_build_id),
                'urn:epic:cfg:can-join_b': True,
                'urn:epic:cfg:chat-enabled_b': _chat_enabled,
                'urn:epic:cfg:invite-perm_s': 'Noone',
                'urn:epic:cfg:join-request-action_s': 'Manual',
                'urn:epic:cfg:not-accepting-members-reason_i': 0,
                'urn:epic:cfg:party-type-id_s': 'default',
                'urn:epic:cfg:presence-perm_s': 'Noone',
            }
        }

        r = PartyService('/party/api/v1/Fortnite/parties')
        return await self.post(r, json=payload, **kwargs)

    async def party_update_member_meta(self, party_id: str,
                                       user_id: str,
                                       updated_meta: dict,
                                       deleted_meta: list,
                                       overridden_meta: dict,
                                       revision: int,
                                       override: dict = {},
                                       **kwargs: Any) -> Any:
        payload = {
            'delete': deleted_meta,
            'update': updated_meta,
            'override': overridden_meta,
            'revision': revision,
        }

        r = PartyService(
            ('/party/api/v1/Fortnite/parties/{party_id}/members/'
             '{user_id}/meta'),
            party_id=party_id,
            user_id=user_id
        )
        return await self.patch(r, json=payload, **kwargs)

    async def party_update_meta(self, party_id: str,
                                updated_meta: dict,
                                deleted_meta: list,
                                overridden_meta: dict,
                                revision: int,
                                config: dict = {},
                                **kwargs: Any) -> Any:
        payload = {
            'meta': {
                'delete': deleted_meta,
                'update': updated_meta,
                'override': overridden_meta
            },
            'revision': revision,
        }

        if config:
            payload['config'] = config

        r = PartyService('/party/api/v1/Fortnite/parties/{party_id}',
                         party_id=party_id)
        return await self.patch(r, json=payload, **kwargs)



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2021 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import asyncio
import logging
import uuid
import time

from aioconsole import ainput
from typing import TYPE_CHECKING, Optional, Any, List

from .errors import AuthException, HTTPException
from .typedefs import StrOrMaybeCoro

if TYPE_CHECKING:
    from .client import Client

log = logging.getLogger(__name__)
_prompt_lock = asyncio.Lock()


class Auth:
    def __init__(self, **kwargs: Any) -> None:
        self.ios_token = kwargs.get('ios_token', 'MzQ0NmNkNzI2OTRjNGE0NDg1ZDgxYjc3YWRiYjIxNDE6OTIwOWQ0YTVlMjVhNDU3ZmI5YjA3NDg5ZDMxM2I0MWE=')  # noqa
        self.fortnite_token = kwargs.get('fortnite_token', 'ZWM2ODRiOGM2ODdmNDc5ZmFkZWEzY2IyYWQ4M2Y1YzY6ZTFmMzFjMjExZjI4NDEzMTg2MjYyZDM3YTEzZmM4NGQ=')  # noqa

    def initialize(self, client: 'Client') -> None:
        self.client = client
        self.device_id = getattr(self, 'device_id', None) or uuid.uuid4().hex
        self._refresh_event = asyncio.Event()
        self._refresh_lock = asyncio.Lock()
        self.refresh_i = 0

    @property
    def ios_authorization(self) -> str:
        return 'bearer {0}'.format(self.ios_access_token)

    @property
    def authorization(self) -> str:
        return 'bearer {0}'.format(self.access_token)

    @property
    def identifier(self) -> str:
        raise NotImplementedError

    def eula_check_needed(self) -> bool:
        return True

    async def authenticate(self, **kwargs) -> dict:
        raise NotImplementedError

    async def _authenticate(self, priority: int = 0) -> None:
        max_attempts = 3
        for i in range(max_attempts):
            try:
                log.info('Running authentication.')
                return await self.authenticate(priority=priority)
            except HTTPException as exc:
                codes = (
                    ('errors.com.epicgames.account.oauth.'
                     'exchange_code_not_found'),
                    ('errors.com.epicgames.account.oauth.'
                     'expired_exchange_code_session'),
                )
                if exc.message_code in codes:
                    if i != max_attempts-1:
                        continue

                raise
            except asyncio.CancelledError:
                return False

    async def reauthenticate(self, priority: int = 0) -> dict:
        raise NotImplementedError

    async def get_eula_version(self, **kwargs: Any) -> int:
        data = await self.client.http.eulatracking_get_data(**kwargs)
        return data['version'] if isinstance(data, dict) else 0

    async def accept_eula(self, **kwargs: Any) -> None:
        version = await self.get_eula_version(**kwargs)
        if version != 0:
            await self.client.http.eulatracking_accept(
                version,
                **kwargs
            )

            try:
                await self.client.http.fortnite_grant_access(**kwargs)
            except HTTPException as e:
                if e.message_code != 'errors.com.epicgames.bad_request':
                    raise

    def _update_ios_data(self, data: dict) -> None:
        self.ios_access_token = data['access_token']
        self.ios_expires_in = data['expires_in']
        self.ios_expires_at = self.client.from_iso(data["expires_at"])
        self.ios_token_type = data['token_type']
        self.ios_refresh_token = data['refresh_token']
        self.ios_refresh_expires = data['refresh_expires']
        self.ios_refresh_expires_at = data['refresh_expires_at']
        self.ios_account_id = data['account_id']
        self.ios_client_id = data['client_id']
        self.ios_internal_client = data['internal_client']
        self.ios_client_service = data['client_service']
        self.ios_app = data['app']
        self.ios_in_app_id = data['in_app_id']

    def _update_data(self, data: dict) -> None:
        self.access_token = data['access_token']
        self.expires_in = data['expires_in']
        self.expires_at = self.client.from_iso(data["expires_at"])
        self.token_type = data['token_type']
        self.refresh_token = data['refresh_token']
        self.refresh_expires = data['refresh_expires']
        self.refresh_expires_at = data['refresh_expires_at']
        self.account_id = data['account_id']
        self.client_id = data['client_id']
        self.internal_client = data['internal_client']
        self.client_service = data['client_service']
        self.app = data['app']
        self.in_app_id = data['in_app_id']

    async def grant_refresh_token(self, refresh_token: str, auth_token: str, *,
                                  priority: int = 0) -> dict:
        payload = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }

        return await self.client.http.account_oauth_grant(
            auth='basic {0}'.format(auth_token),
            device_id=True,
            data=payload,
            priority=priority
        )

    async def get_exchange_code(self, *,
                                auth='IOS_ACCESS_TOKEN',
                                priority: int = 0) -> str:
        data = await self.client.http.account_get_exchange_data(
            auth=auth,
            priority=priority
        )
        return data['code']

    async def exchange_code_for_session(self, token: str, code: str, *,
                                        priority: int = 0) -> dict:
        payload = {
            'grant_type': 'exchange_code',
            'exchange_code': code,
            'token_type': 'eg1',
        }

        return await self.client.http.account_oauth_grant(
            auth='basic {0}'.format(token),
            device_id=True,
            data=payload,
            priority=priority
        )

    async def kill_token(self, token: str) -> None:
        await self.client.http.account_sessions_kill_token(
            token,
            auth='bearer {0}'.format(token)
        )

    async def kill_other_sessions(self, auth: str = 'IOS_ACCESS_TOKEN', *,
                                  priority: int = 0) -> None:
        await self.client.http.account_sessions_kill(
            'OTHERS_ACCOUNT_CLIENT_SERVICE',
            auth=auth,
            priority=priority
        )
        log.debug('Killing other sessions')

    def refresh_loop_running(self):
        task = self.client._refresh_task
        return task is not None and not task.cancelled()

    async def schedule_token_refresh(self) -> None:
        subtracted = self.ios_expires_at - datetime.datetime.utcnow()
        self.token_timeout = (subtracted).total_seconds() - 300
        await asyncio.sleep(self.token_timeout)

    async def run_refresh_loop(self) -> None:
        loop = self.client.loop

        while True:
            self._refresh_event.clear()
            _, p = await asyncio.wait((
                loop.create_task(self._refresh_event.wait()),
                loop.create_task(self.schedule_token_refresh())
            ), return_when=asyncio.FIRST_COMPLETED)

            for pending in p:
                if not pending.cancelled():
                    pending.cancel()

            await self.do_refresh()

    async def do_refresh(self) -> None:
        reauth_lock = self.client._reauth_lock
        reauth_lock.priority += 1
        forced = reauth_lock.locked()

        try:
            if not forced:
                await self._refresh_lock.acquire()

            log.debug('Refreshing session')
            self.client._refresh_times.append(time.time())

            try:
                data = await self.grant_refresh_token(
                    self.ios_refresh_token,
                    self.ios_token,
                    priority=reauth_lock.priority
                )
                self._update_ios_data(data)

                data = await self.grant_refresh_token(
                    self.refresh_token,
                    self.fortnite_token,
                    priority=reauth_lock.priority
                )
                self._update_data(data)
            except (HTTPException, AttributeError) as exc:
                m = 'errors.com.epicgames.account.auth_token.' \
                    'invalid_refresh_token'
                if isinstance(exc, HTTPException) and exc.message_code != m:
                    raise

                log.debug(
                    'Invalid refresh token was supplied. Attempting to '
                    'reconnect if possible.'
                )
                try:
                    await self.reauthenticate(
                        priority=reauth_lock.priority
                    )
                except NotImplementedError:
                    raise exc

                log.debug('Successfully reauthenticated.')

            try:
                log.debug('Refreshing xmpp session')
                await self.client.xmpp.close()
                await self.client.xmpp.run()

                await self.client._reconnect_to_party()
            except AttributeError:
                pass

            self.refresh_i += 1
            log.debug('Sessions was successfully refreshed.')
            self.client.dispatch_event('auth_refresh')

        finally:
            if not forced:
                self._refresh_lock.release()

    async def run_refresh(self) -> None:
        self._refresh_event.set()
        await self.client.wait_for('auth_refresh')

    def refreshing(self) -> bool:
        return self._refresh_lock.locked()

    async def fetch_device_auths(self) -> List[dict]:
        # Return payload:
        # [
        #     {
        #         "deviceId": "01fbc14162634294a1db59ddced3c10c",
        #         "accountId": "4a6313cbbe7d432b8132b5ee67ad53fa",
        #         "userAgent": "EpicGamesLauncher/++Fortnite+Release-12.00-CL-11586896 Windows/10.0.17134.1.768.64bit",  # noqa
        #         "created": {
        #             "location": "Oslo, Norway",
        #             "ipAddress": "",  # ipv4 address
        #             "dateTime": "2020-04-25T20:38:57.570Z"
        #         },
        #         "lastAccess": {
        #             "location": "Oslo, Norway",
        #             "ipAddress": "",  # ipv4 address
        #             "dateTime": "2020-05-13T16:33:43.075Z"
        #         }
        #     }
        # ]
        return await self.client.http.account_get_device_auths(
            self.ios_account_id
        )

    async def generate_device_auth(self) -> dict:
        # Return payload:
        # {
        #     "deviceId": "e2ba6aa72411468ba4fee016086809d7",
        #     "accountId": "4a6313cbbe7d432b8132b5ee67ad53fa",
        #     "secret": "",  # 32 char (not hex)
        #     "userAgent": "EpicGamesLauncher/++Fortnite+Release-12.00-CL-11586896 Windows/10.0.17134.1.768.64bit",  # noqa
        #     "created": {
        #         "location": "Oslo, Norway",
        #         "ipAddress": "",  # ipv4
        #         "dateTime": "2020-05-13T18:32:07.601Z"
        #     }
        # }
        return await self.client.http.account_generate_device_auth(
            self.ios_account_id
        )

    async def delete_device_auth(self, device_id: str) -> None:
        await self.client.http.account_delete_device_auth(
            self.ios_account_id,
            device_id
        )


class EmailAndPasswordAuth(Auth):
    """Authenticates by email and password.

    .. warning::

        Some users might experience an error saying captcha was invalid.
        If this is the case, use :class:`AdvancedAuth` with an exchange code
        to generate a device auth.

    Parameters
    ----------
    email: :class:`str`
        The accounts email.
    password: :class:`str`
        The accounts password.
    two_factor_code: Optional[:class:`int`]
        The current two factor code if needed. If not passed here, it
        will be prompted later.
    device_id: Optional[:class:`str`]
        A 32 char hex representing your device.
    ios_token: Optional[:class:`str`]
        The ios token to use with authentication. You should generally
        not need to set this manually.
    fortnite_token: Optional[:class:`str`]
        The fortnite token to use with authentication. You should generally
        not need to set this manually.
    """
    def __init__(self, email: str, password: str, *,
                 two_factor_code: Optional[int] = None,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.email = email
        self.password = password
        self.two_factor_code = two_factor_code

    @property
    def identifier(self) -> str:
        return self.email

    async def fetch_xsrf_token(self) -> str:
        response = await self.client.http.epicgames_get_csrf()
        return response.cookies['XSRF-TOKEN'].value

    async def ios_authenticate(self) -> dict:
        log.info('Fetching valid xsrf token.')
        token = await self.fetch_xsrf_token()

        await self.client.http.epicgames_reputation(token)

        try:
            log.info('Logging in.')
            await self.client.http.epicgames_login(
                self.email,
                self.password,
                token
            )
        except HTTPException as e:
            m = 'errors.com.epicgames.account.invalid_account_credentials'
            if e.message_code == m:
                raise AuthException(
                    'Invalid account credentials passed.',
                    e
                ) from e

            if e.message_code != ('errors.com.epicgames.common.'
                                  'two_factor_authentication.required'):
                raise

            log.info('Logging in interrupted. 2fa required.')
            log.info('Fetching new valid xsrf token.')
            token = await self.fetch_xsrf_token()

            code = self.two_factor_code
            if code is None:
                async with _prompt_lock:
                    code = await ainput(
                        'Please enter the 2fa code:\n',
                        loop=self.client.loop
                    )

            try:
                await self.client.http.epicgames_mfa_login(
                    e.raw['metadata']['twoFactorMethod'],
                    code,
                    token
                )
            except HTTPException as exc:
                m = (
                    'errors.com.epicgames.accountportal.mfa_code_invalid',
                    'errors.com.epicgames.accountportal.validation'
                )
                if exc.message_code in m:
                    raise AuthException(
                        'Invalid 2fa code passed.',
                        exc
                    ) from exc

                raise

        await self.client.http.epicgames_redirect(token)

        token = await self.fetch_xsrf_token()
        log.info('Fetching exchange code.')
        data = await self.client.http.epicgames_get_exchange_data(token)

        log.info('Exchanging code.')
        data = await self.exchange_code_for_session(
            self.ios_token,
            data['code']
        )
        return data

    async def authenticate(self, **kwargs) -> None:
        data = await self.ios_authenticate()
        self._update_ios_data(data)

        if self.client.kill_other_sessions:
            await self.kill_other_sessions()

        code = await self.get_exchange_code()
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code
        )
        self._update_data(data)


class ExchangeCodeAuth(Auth):
    """Authenticates by an exchange code.

    .. note::

        The method to get an exchange code has been significantly harder
        since epic patched the old method of copying the code from one of
        their own endpoints that could be requested easily in a browser.
        To obtain an exchange code it is recommended to provide a custom
        solution like running a selenium process where you log in on
        https://epicgames.com and then redirect to /id/api/exchange/generate.
        You can then return the exchange code. You can put this solution
        in a function and then pass this to ``exchange_code``.

    .. note::

        An exchange code only works for a single login within a short
        timeframe (300 seconds). Therefore you need to get a new code for each
        login. You can get a new code by refreshing the site.

    Parameters
    ----------
    code: Union[:class:`str`, Callable, Awaitable]
        The exchange code or a function/coroutine that when called returns
        the exchange code.
    device_id: Optional[:class:`str`]
        A 32 char hex string representing your device.
    ios_token: Optional[:class:`str`]
        The ios token to use with authentication. You should generally
        not need to set this manually.
    fortnite_token: Optional[:class:`str`]
        The fortnite token to use with authentication. You should generally
        not need to set this manually.
    """
    def __init__(self, code: StrOrMaybeCoro,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.code = code
        self.resolved_code = None

    async def resolve(self, code: StrOrMaybeCoro) -> str:
        if isinstance(code, str):
            return code

        elif asyncio.iscoroutinefunction(code):
            res = await code()
        else:
            res = code()

        if not isinstance(res, str):
            raise TypeError('Return type of callable func/coro must be str')

        return res

    @property
    def identifier(self) -> str:
        return self.resolved_code

    async def ios_authenticate(self) -> dict:
        log.info('Exchanging code.')
        self.resolved_code = await self.resolve(self.code)

        try:
            data = await self.exchange_code_for_session(
                self.ios_token,
                self.resolved_code
            )
        except HTTPException as e:
            m = 'errors.com.epicgames.account.oauth.exchange_code_not_found'
            if e.message_code == m:
                raise AuthException(
                    'Invalid exchange code supplied',
                    e
                ) from e

            raise

        return data

    async def authenticate(self, **kwargs) -> None:
        data = await self.ios_authenticate()
        self._update_ios_data(data)

        if self.client.kill_other_sessions:
            await self.kill_other_sessions()

        code = await self.get_exchange_code()
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code
        )
        self._update_data(data)


class AuthorizationCodeAuth(ExchangeCodeAuth):
    """Authenticates by exchange code.

    You can get the code from `here
    <https://www.epicgames.com/id/api/redirect?
    clientId=3446cd72694c4a4485d81b77adbb2141&responseType=code>`_ by logging
    in and copying the code from the redirectUrl's query parameters. If you
    are already logged in and want to change accounts, simply log out at
    https://www.epicgames.com, log in to the new account and then enter the
    link above again to generate an authorization code.

    .. note::

        An authorization code only works for a single login within a short
        timeframe (300 seconds). Therefore you need to get a new code for each
        login. You can get a new code by refreshing the site.

    Parameters
    ----------
    code: Union[:class:`str`, Callable, Awaitable]
        The authorization code or a function/coroutine that when called returns
        the authorization code.
    device_id: Optional[:class:`str`]
        A 32 char hex string representing your device.
    ios_token: Optional[:class:`str`]
        The ios token to use with authentication. You should generally
        not need to set this manually.
    fortnite_token: Optional[:class:`str`]
        The fortnite token to use with authentication. You should generally
        not need to set this manually.
    """
    def __init__(self, code: StrOrMaybeCoro,
                 **kwargs: Any) -> None:
        super().__init__(code, **kwargs)

    async def ios_authenticate(self) -> dict:
        self.resolved_code = await self.resolve(self.code)

        payload = {
            'grant_type': 'authorization_code',
            'code': self.resolved_code,
        }

        try:
            data = await self.client.http.account_oauth_grant(
                auth='basic {0}'.format(self.ios_token),
                device_id=True,
                data=payload
            )
        except HTTPException as e:
            m = 'errors.com.epicgames.account.oauth.authorization_code_not_found'  # noqa
            if e.message_code == m:
                raise AuthException(
                    'Invalid authorization code supplied',
                    e
                ) from e

            raise

        return data


class DeviceAuth(Auth):
    """Authenticate with device auth details.

    .. note::

        All device auths generated for an account is removed once the accounts
        password gets reset. If you managed to leak you device_id and secret,
        simply reset the accounts password and everything should be fine.

    Parameters
    ----------
    device_id: :class:`str`
        The device id.
    account_id: :class:`str`
        The account's id.
    secret: :class:`str`
        The secret.
    ios_token: Optional[:class:`str`]
        The ios token to use with authentication. You should generally
        not need to set this manually.
    fortnite_token: Optional[:class:`str`]
        The fortnite token to use with authentication. You should generally
        not need to set this manually.
    """
    def __init__(self, device_id: str,
                 account_id: str,
                 secret: str,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.device_id = device_id
        self.account_id = account_id
        self.secret = secret

    @property
    def identifier(self) -> str:
        return self.account_id

    def eula_check_needed(self) -> bool:
        return False

    async def ios_authenticate(self, priority: int = 0) -> dict:
        payload = {
            'grant_type': 'device_auth',
            'device_id': self.device_id,
            'account_id': self.account_id,
            'secret': self.secret,
            'token_type': 'eg1'
        }

        try:
            data = await self.client.http.account_oauth_grant(
                auth='basic {0}'.format(self.ios_token),
                data=payload,
                priority=priority
            )
        except HTTPException as exc:
            m = 'errors.com.epicgames.account.invalid_account_credentials'
            if exc.message_code == m:
                raise AuthException(
                    'Invalid device auth details passed.',
                    exc
                ) from exc

            raise

        return data

    async def authenticate(self, priority: int = 0) -> None:
        data = await self.ios_authenticate(priority=priority)
        self._update_ios_data(data)

        if self.client.kill_other_sessions:
            await self.kill_other_sessions(priority=priority)

        code = await self.get_exchange_code(priority=priority)
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code,
            priority=priority
        )
        self._update_data(data)

    async def reauthenticate(self, priority: int = 0) -> None:
        """Used for reauthenticating if refreshing fails."""
        log.debug('Starting reauthentication.')

        ret = await self.authenticate(priority=priority)
        log.debug('Successfully reauthenicated.')
        return ret


class RefreshTokenAuth(Auth):
    """Authenticates by the passed launcher refresh token.

    Parameters
    ----------
    refresh_token: :class:`str`
        A valid launcher refresh token.
    """
    def __init__(self, refresh_token: str,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self._refresh_token = refresh_token

    @property
    def identifier(self) -> str:
        return self._refresh_token

    def eula_check_needed(self) -> bool:
        return False

    async def ios_authenticate(self, priority: int = 0) -> dict:
        data = await self.grant_refresh_token(
            self._refresh_token,
            self.ios_token,
            priority=priority
        )
        return data

    async def authenticate(self, priority: int = 0) -> None:
        data = await self.ios_authenticate(priority=priority)
        self._update_ios_data(data)

        code = await self.get_exchange_code(priority=priority)
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code,
            priority=priority
        )
        self._update_data(data)


class AdvancedAuth(Auth):
    """Authenticates by the available data in the following order:

    1. By :class:`DeviceAuth` if ``device_id``, ``account_id`` and ``secret``
    are present.
    2. By :class:`EmailAndPasswordAuth` if ``email`` and ``password`` is
    present. If authentication fails because of required captcha, it then
    attempts to authenticate with the next step.
    3. :class:`ExchangeCodeAuth` is tried if ``exchange_code`` is present
    or if ``prompt_exchange_code`` is ``True``.
    4. :class:`AuthorizationCodeAuth` is tried if ``authorization_code`` is
    present or if ``prompt_authorization_code`` is ``True``.

    If the authentication was not done by step 1, a device auth is
    automatically generated and the details will be dispatched to
    :func:`event_device_auth_generate`. It is important to store
    these values somewhere since they can be used for easier logins.

    Parameters
    ----------
    email: Optional[:class:`str`]
        The email to use for the login.
    password: Optional[:class:`str`]
        The password to use for the login.
    two_factor_code: Optional[:class:`int`]
        The two factor code to use for the login if needed. If this is
        not passed but later needed, you will be prompted to enter it
        in the console.
    exchange_code: Optional[Union[:class:`str`, Callable, Awaitable]]
        The exchange code or a function/coroutine that when called returns
        the exchange code.
    authorization_code: Optional[Union[:class:`str`, Callable, Awaitable]]
        The authorization code or a function/coroutine that when called returns
        the authorization code.
    device_id: Optional[:class:`str`]
        The device id to use for the login.
    account_id: Optional[:class:`str`]
        The account id to use for the login.
    secret: Optional[:class:`str`]
        The secret to use for the login.
    prompt_exchange_code: :class:`bool`
        If this is set to ``True`` and no exchange code is passed,
        you will be prompted to enter the exchange code in the console
        if needed.

        .. note::

            Both ``prompt_exchange_code`` and ``prompt_authorization_code``
            cannot be True at the same time.
    prompt_authorization_code: :class:`bool`
        If this is set to ``True`` and no authorization code is passed,
        you will be prompted to enter the authorization code in the console
        if needed.

        .. note::

            Both ``prompt_exchange_code`` and ``prompt_authorization_code``
            cannot be True at the same time.
    prompt_code_if_invalid: :class:`bool`
        Whether or not to prompt a code if the device auth details
        was invalid. If this is False then the regular :exc:`AuthException` is
        raised instead.

        .. note::

            This only works if ``prompt_exchange_code`` or
            ``prompt_authorization_code`` is ``True``.
    prompt_code_if_throttled: :class:`bool`
        If this is set to ``True`` and you receive a throttling response,
        you will be prompted to enter a code in the console.

        .. note::

            This only works if ``prompt_exchange_code`` or
            ``prompt_authorization_code`` is ``True``.
    delete_existing_device_auths: :class:`bool`
        Whether or not to delete all existing device auths when a new
        is created.
    ios_token: Optional[:class:`str`]
        The ios token to use with authentication. You should generally
        not need to set this manually.
    fortnite_token: Optional[:class:`str`]
        The fortnite token to use with authentication. You should generally
        not need to set this manually.
    """
    def __init__(self, email: Optional[str] = None,
                 password: Optional[str] = None,
                 two_factor_code: Optional[int] = None,
                 exchange_code: Optional[StrOrMaybeCoro] = None,
                 authorization_code: Optional[StrOrMaybeCoro] = None,
                 device_id: Optional[str] = None,
                 account_id: Optional[str] = None,
                 secret: Optional[str] = None,
                 prompt_exchange_code: bool = False,
                 prompt_authorization_code: bool = False,
                 prompt_code_if_invalid: bool = False,
                 prompt_code_if_throttled: bool = False,
                 delete_existing_device_auths: bool = False,
                 **kwargs: Any) -> None:
        super().__init__(**kwargs)

        self.email = email
        self.password = password
        self.two_factor_code = two_factor_code
        self.exchange_code = exchange_code
        self.authorization_code = authorization_code
        self.device_id = device_id
        self.account_id = account_id
        self.secret = secret

        self.delete_existing_device_auths = delete_existing_device_auths
        self.prompt_exchange_code = prompt_exchange_code
        self.prompt_authorization_code = prompt_authorization_code

        if self.prompt_exchange_code and self.prompt_authorization_code:
            raise ValueError('Both prompt_exchange_code and '
                             'prompt_authorization_code cannot be True at '
                             'the same time.')

        self.prompt_code_if_invalid = prompt_code_if_invalid
        self.prompt_code_if_throttled = prompt_code_if_throttled
        self.kwargs = kwargs

        self._used_auth = None

    @property
    def identifier(self) -> str:
        return self.email or self.account_id or self.exchange_code

    def eula_check_needed(self) -> bool:
        return self._used_auth.eula_check_needed()

    def email_and_password_ready(self) -> bool:
        return self.email and self.password

    def exchange_code_ready(self) -> bool:
        return self.exchange_code is not None

    def authorization_code_ready(self) -> bool:
        return self.authorization_code is not None

    def code_ready(self) -> bool:
        return self.exchange_code_ready() or self.authorization_code_ready()

    def device_auth_ready(self) -> bool:
        return self.device_id and self.account_id and self.secret

    def prompt_enabled(self) -> bool:
        return True in (self.prompt_exchange_code, self.prompt_authorization_code)  # noqa

    def get_prompt_type_name(self) -> str:
        if self.prompt_exchange_code:
            return 'exchange'
        elif self.prompt_authorization_code:
            return 'authorization'

    async def run_email_and_password_authenticate(self) -> dict:
        auth = EmailAndPasswordAuth(
            email=self.email,
            password=self.password,
            two_factor_code=self.two_factor_code,
            **self.kwargs
        )
        auth.initialize(self.client)
        self._used_auth = auth

        return await auth.ios_authenticate()

    async def run_exchange_code_authenticate(self) -> dict:
        auth = ExchangeCodeAuth(
            code=self.exchange_code,
            **self.kwargs
        )
        auth.initialize(self.client)
        self._used_auth = auth

        return await auth.ios_authenticate()

    async def run_authorization_code_authenticate(self) -> dict:
        auth = AuthorizationCodeAuth(
            code=self.authorization_code,
            **self.kwargs
        )
        auth.initialize(self.client)
        self._used_auth = auth

        return await auth.ios_authenticate()

    async def run_device_authenticate(self, device_id: Optional[str] = None,
                                      account_id: Optional[str] = None,
                                      secret: Optional[str] = None,
                                      *, priority: int = 0
                                      ) -> dict:
        auth = DeviceAuth(
            device_id=device_id or self.device_id,
            account_id=account_id or self.account_id,
            secret=secret or self.secret,
            **self.kwargs
        )
        auth.initialize(self.client)
        self._used_auth = auth

        data = await auth.ios_authenticate(priority=priority)
        self._update_ios_data(data)

        code = await auth.get_exchange_code(priority=priority)
        return await auth.exchange_code_for_session(
            auth.fortnite_token,
            code,
            priority=priority
        )

    async def ios_authenticate(self) -> dict:
        data = None
        prompt_message = ''

        if self.device_auth_ready():
            try:
                return await self.run_device_authenticate()
            except AuthException as exc:
                original = exc.original
                if not self.prompt_enabled() or not self.prompt_code_if_invalid:  # noqa
                    raise

                if isinstance(original, HTTPException):
                    m = 'errors.com.epicgames.account.invalid_account_credentials'  # noqa
                    if original.message_code != m:
                        raise

                prompt_message = 'Invalid device auth details passed. '

        elif self.email_and_password_ready():
            try:
                data = await self.run_email_and_password_authenticate()
            except HTTPException as e:
                m = {
                    'errors.com.epicgames.accountportal.captcha_invalid': 'Captcha was enforced. '  # noqa
                }
                if self.prompt_enabled() and self.prompt_code_if_throttled:
                    m['errors.com.epicgames.common.throttled'] = 'Account was throttled. '  # noqa

                if e.message_code not in m:
                    raise

                short = m.get(e.message_code)
                if (short is not None
                        and not self.code_ready()
                        and not self.prompt_enabled()):
                    raise AuthException(
                        'This account requires an exchange or authorization '
                        'code.',
                        e
                    ) from e

                prompt_message = short

        if data is None:
            prompted = False
            code = None
            if not self.code_ready() and self.prompt_enabled():
                prompted = True
                code_type = self.get_prompt_type_name()
                if self.email is not None:
                    text = '{0}Please enter a valid {1} code ' \
                           'for {2}\n'.format(
                                prompt_message,
                                code_type,
                                self.email
                            )
                else:
                    text = '{0}Please enter a valid {1} code.\n'.format(
                        prompt_message,
                        code_type
                    )

                async with _prompt_lock:
                    code = await ainput(
                        text,
                        loop=self.client.loop
                    )

            if (prompted and self.prompt_exchange_code) or self.exchange_code_ready():  # noqa
                self.exchange_code = code or self.exchange_code
                data = await self.run_exchange_code_authenticate()
            else:
                self.authorization_code = code or self.authorization_code
                data = await self.run_authorization_code_authenticate()

        self._update_ios_data(data)

        if self.delete_existing_device_auths:
            tasks = []
            auths = await self.fetch_device_auths()
            for auth in auths:
                tasks.append(self.client.loop.create_task(
                    self.delete_device_auth(
                        auth['deviceId']
                    )
                ))

            if tasks:
                await asyncio.gather(*tasks)

        data = await self.generate_device_auth()
        details = {
            'device_id': data['deviceId'],
            'account_id': data['accountId'],
            'secret': data['secret'],
        }
        self.__dict__.update(details)

        account_data = await self.client.http.account_get_by_user_id(
            self.ios_account_id,
            auth='IOS_ACCESS_TOKEN'
        )

        self.client.dispatch_event(
            'device_auth_generate',
            details,
            account_data['email']
        )

        return data

    async def authenticate(self, **kwargs) -> None:
        await self.ios_authenticate()

        if self.client.kill_other_sessions:
            await self.kill_other_sessions()

        code = await self.get_exchange_code()
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code
        )
        self._update_data(data)

    async def reauthenticate(self, priority: int = 0) -> None:
        log.debug('Starting reauthentication.')

        await self.run_device_authenticate(
            device_id=self.device_id,
            account_id=self.account_id,
            secret=self.secret,
            priority=priority,
        )

        if self.client.kill_other_sessions:
            await self.kill_other_sessions(priority=priority)

        code = await self.get_exchange_code(priority=priority)
        data = await self.exchange_code_for_session(
            self.fortnite_token,
            code,
            priority=priority
        )
        self._update_data(data)
        log.debug('Successfully reauthenticated.')



"""
MIT License

Copyright (c) 2019-2021 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json

from typing import List, Optional, Dict, Any, Union
from .enums import DefaultCharactersChapter1, KairosBackgroundColorPreset

ListOrPreset = Optional[Union[List[str], KairosBackgroundColorPreset]]


def get_random_default_avatar() -> 'Avatar':
    return Avatar(
        asset=DefaultCharactersChapter1.get_random_name(),
        background_colors=KairosBackgroundColorPreset.get_random_value()
    )


class Avatar:
    """Dataclass which represents a kairos avatar.

    Parameters
    ----------
    asset: Optional[:class:`str`]
        The CID to use as the asset.
    background_colors: Optional[List[:class:`str`]]
        A list of exactly three hex color values represented as strings.
        In Kairos (PartyHub) these values will be used to create a gradiant
        background. Fortnite however will find the average
        of the three colors and use that.

    Attributes
    ----------
    asset: Optional[:class:`str`]
        The CID used for this asset.
    background: Optional[Union[List[:class:`str`], :class:`KairosBackgroundColorPreset`]]
        A list of exactly three hex color values represented as strings.
    """  # noqa

    __slots__ = ('asset', 'background_colors')

    def __init__(self, *, asset: Optional[str] = None,
                 background_colors: ListOrPreset = None) -> None:
        self.asset = asset

        if isinstance(background_colors, KairosBackgroundColorPreset):
            self.background_colors = background_colors.value
        else:
            self.background_colors = background_colors

    def __repr__(self) -> str:
        return ('<Avatar asset={0.asset!r} '
                'background_colors={0.background_colors!r}>'.format(self))

    def __eq__(self, other):
        return isinstance(other, Avatar) and other.to_dict() == self.to_dict()

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> Dict[str, Any]:
        """Converts it into a fortnite friendly dict.

        Returns
        -------
        Dict[:class:`str`, Any]
            The values transformed into a fortnite friendly dictionary.
        """
        data = {'appInstalled': 'init'}

        if self.asset:
            data['avatar'] = self.asset.lower()

        if self.background_colors:
            colors = [c.upper() for c in self.background_colors]
            dumped = json.dumps(colors)

            # For some reason we have to remove the spaces or else it won't
            # show correctly client sided.
            data['avatarBackground'] = dumped.replace(' ', '')

        return data



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import inspect
import asyncio

from typing import (TYPE_CHECKING, Union, Awaitable, Any, List, Dict, Iterable,
                    Optional)
from fortnitepy.typedefs import MaybeCoro

from ._types import _BaseCommand
from .context import Context

if TYPE_CHECKING:
    from .bot import Bot


__all__ = (
    'CogMeta',
    'Cog',
)


class CogMeta(type):
    """A metaclass for defining a cog.

    Note that you should probably not use this directly. It is exposed
    purely for documentation purposes.

    .. note::

        When passing an attribute of a metaclass that is documented below, note
        that you must pass it as a keyword-only argument to the class creation
        like the following example:

        .. code-block:: python3

            class MyCog(commands.Cog, name='My Cog'):
                pass

    Attributes
    -----------
    name: :class:`str`
        The cog name. By default, it is the name of the class with no
        modification.
    command_attrs: :class:`dict`
        A list of attributes to apply to every command inside this cog.
        The dictionary is passed into the :class:`Command` (or its subclass)
        options at ``__init__``. If you specify attributes inside the command
        attribute in the class, it will override the one specified inside this
        attribute. For example:

        .. code-block:: python3

            class MyCog(commands.Cog, command_attrs=dict(hidden=True)):

                @commands.command()
                async def foo(self, ctx):
                    pass # hidden -> True

                @commands.command(hidden=False)
                async def bar(self, ctx):
                    pass # hidden -> False
    """

    def __new__(cls, *args: list, **kwargs: dict) -> None:
        name, bases, attrs = args
        attrs['__cog_name__'] = cog_name = kwargs.pop('name', name)

        last = cog_name.split()[-1]
        try:
            int(last)
        except ValueError:
            pass
        else:
            raise NameError('Cog names cannot end with a space followed by a '
                            'number or solely consist of a number.')

        command_attrs = kwargs.pop('command_attrs', {})
        attrs['__cog_settings__'] = command_attrs

        commands = {}
        event_handlers = {}
        no_bot_cog = ('Commands or listeners must not start with cog_ or bot_ '
                      '(in method {0.__name__}.{1})')

        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        for base in reversed(new_cls.__mro__):
            for elem, value in base.__dict__.items():
                if elem in commands:
                    del commands[elem]
                if elem in event_handlers:
                    del event_handlers[elem]

                is_static_method = isinstance(value, staticmethod)
                if is_static_method:
                    value = value.__func__

                if isinstance(value, _BaseCommand):
                    if is_static_method:
                        raise TypeError(
                            'Command in method {0}.{1!r} must not be '
                            'staticmethod.'.format(base, elem)
                        )

                    commands[elem] = value

                elif asyncio.iscoroutinefunction(value):
                    try:
                        getattr(value, '__cog_event_handler__')
                    except AttributeError:
                        pass
                    else:
                        if elem.startswith(('cog_', 'bot_')):
                            raise TypeError(no_bot_cog.format(base, elem))
                        event_handlers[elem] = value

        new_cls.__cog_commands__ = list(commands.values())

        event_handlers_as_list = []
        for handler in event_handlers.values():
            for handler_name, was_def in handler.__cog_event_handler_names__:
                event_handlers_as_list.append((
                    handler_name,
                    handler.__name__,
                    was_def
                ))

        new_cls.__cog_event_handlers__ = event_handlers_as_list
        return new_cls

    def __init__(self, *args: list, **kwargs: dict) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def qualified_name(cls) -> str:
        return cls.__cog_name__


def _cog_special_method(func: MaybeCoro) -> MaybeCoro:
    func.__cog_special_method__ = None
    return func


class Cog(metaclass=CogMeta):
    """The base class that all cogs must inherit from.

    A cog is a collection of commands, listeners, and optional state to
    help group commands together. More information on them can be found on
    the :ref:`ext_commands_cogs` page.

    When inheriting from this class, the options shown in :class:`CogMeta`
    are equally valid here.
    """

    def __new__(cls, *args: list, **kwargs: dict) -> None:
        self = super().__new__(cls)
        cmd_attrs = cls.__cog_settings__

        self.__cog_commands__ = tuple(
            c._update_copy(cmd_attrs) for c in cls.__cog_commands__
        )

        lookup = {
            cmd.qualified_name: cmd for cmd in self.__cog_commands__
        }

        for command in self.__cog_commands__:
            setattr(self, command.callback.__name__, command)
            parent = command.parent
            if parent is not None:
                parent = lookup[parent.qualified_name]

                parent.remove_command(command.name)
                parent.add_command(command)

        return self

    def get_commands(self) -> List[_BaseCommand]:
        r"""Returns a :class:`list` of :class:`.Command`\s that are
        defined inside this cog.

        .. note::

            This does not include subcommands.
        """
        return [c for c in self.__cog_commands__ if c.parent is None]

    @property
    def qualified_name(self) -> str:
        """:class:`str`: Returns the cog's specified name, not
        the class name.
        """
        return self.__cog_name__

    @property
    def description(self) -> str:
        """:class:`str`: Returns the cog's description, typically
        the cleaned docstring.
        """
        try:
            return self.__cog_cleaned_doc__
        except AttributeError:
            self.__cog_cleaned_doc__ = inspect.getdoc(self)
            return self.__cog_cleaned_doc__

    def walk_commands(self) -> Iterable:
        """An iterator that recursively walks through this cog's
        commands and subcommands.
        """
        from .core import GroupMixin

        for command in self.__cog_commands__:
            if command.parent is None:
                yield command
                if isinstance(command, GroupMixin):
                    yield from command.walk_commands()

    @property
    def event_handlers(self) -> Dict[str, Awaitable]:
        """:class:`list`: A list of (name, function) event handler pairs
        that are defined in this cog.
        """
        return {
            name: getattr(self, method_name)
            for name, method_name, _
            in self.__cog_event_handlers__
        }

    @classmethod
    def _get_overridden_method(cls, method: MaybeCoro) -> MaybeCoro:
        """Return None if the method is not overridden. Otherwise returns the
        overridden method.
        """
        return getattr(method.__func__, '__cog_special_method__', method)

    @classmethod
    def event(cls, event: Union[str, Awaitable[Any]] = None) -> Awaitable:
        """A decorator to register an event.

        Usage: ::

            @commands.Cog.event()
            async def event_friend_message(message):
                await message.reply('Thanks for your message!')

            @commands.Cog.event('friend_message')
            async def my_message_handler(message):
                await message.reply('Thanks for your message!')

        Raises
        ------
        TypeError
            The decorated function is not a coroutine.
        TypeError
            Event is not specified as argument or function name with event
            prefix.
        """
        is_coro = callable(event)

        def pred(coro):
            if isinstance(coro, staticmethod):
                coro = coro.__func__

            if not asyncio.iscoroutinefunction(coro):
                raise TypeError('the decorated function must be a coroutine')

            if is_coro or event is None:
                name = coro.__name__
                name_defined = False
            else:
                name = event
                name_defined = True

            coro.__cog_event_handler__ = True

            try:
                coro.__cog_event_handler_names__.append((name, name_defined))
            except AttributeError:
                coro.__cog_event_handler_names__ = [(name, name_defined)]

            return coro
        return pred(event) if is_coro else pred

    def _inject(self, bot: 'Bot') -> 'Cog':
        cls = self.__class__

        for index, command in enumerate(self.__cog_commands__):
            command.cog = self
            command.instance = bot
            if command.parent is None:
                try:
                    bot.add_command(command)
                except Exception as e:
                    for to_undo in self.__cog_commands__[:index]:
                        bot.remove_command(to_undo)
                    raise e

        if cls.bot_check is not Cog.bot_check:
            bot.add_check(self.bot_check)

        if cls.bot_check_once is not Cog.bot_check_once:
            bot.add_check(self.bot_check_once, call_once=True)

        # We need to do this to remove all prefixes of events registered
        # with the coro's name instead of manually specifying the name.
        new = []
        for name, method, name_was_defined in self.__cog_event_handlers__:
            if not name_was_defined:
                if name.startswith(bot.event_prefix):
                    name = name[len(bot.event_prefix):]

            new.append((name, method, name_was_defined))

        self.__cog_event_handlers__ = new

        for name, method_name, _ in self.__cog_event_handlers__:
            bot.add_event_handler(name, getattr(self, method_name))

        return self

    def _eject(self, bot: 'Bot') -> None:
        cls = self.__class__

        try:
            for command in self.__cog_commands__:
                if command.parent is None:
                    bot.remove_command(command.name)

            for name, method_name, _ in self.__cog_event_handlers__:
                bot.remove_event_handler(name, getattr(self, method_name))

            if cls.bot_check is not Cog.bot_check:
                bot.remove_check(self.bot_check)

            if cls.bot_check_once is not Cog.bot_check_once:
                bot.remove_check(self.bot_check_once, call_once=True)

        finally:
            self.cog_unload()

    @_cog_special_method
    def cog_unload(self) -> None:
        """A special method that is called when the cog gets removed.

        This function **cannot** be a coroutine. It must be a regular
        function.

        Subclasses must replace this if they want special unloading behaviour.
        """
        pass

    @_cog_special_method
    def bot_check_once(self, ctx: 'Context') -> bool:
        """A special method that registers as a :meth:`.Bot.check_once`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def bot_check(self, ctx: 'Context') -> bool:
        """A special method that registers as a :meth:`.Bot.check`
        check.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_check(self, ctx: 'Context') -> bool:
        """A special method that registers as a :func:`commands.check`
        for every command and subcommand in this cog.

        This function **can** be a coroutine and must take a sole parameter,
        ``ctx``, to represent the :class:`.Context`.
        """
        return True

    @_cog_special_method
    def cog_command_error(self, ctx: 'Context',
                          error: Exception) -> Optional[bool]:
        """A special method that is called whenever an error
        is dispatched inside this cog.

        This is similar to :func:`.event_command_error` except only applying
        to the commands inside this cog.

        Command error handlers are raised in a specific order. Returning
        ``False`` in any of them will invoke the next handler in the chain. If
        there are no handlers left to call, the error is printed.

        The order goes as follows:
        1. The local command error handler is called. (Handler specified by
        :meth:`.Command.check`)
        2. The local cog command error handler is called. (This)
        3. All :func:`.event_command_error()` handlers are called
        simultaneously. If any of them return False, the error will
        be printed.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context where the error happened.
        error: :class:`CommandError`
            The error that happened.
        """
        return False

    @_cog_special_method
    async def cog_before_invoke(self, ctx: 'Context') -> None:
        """A special method that acts as a cog local pre-invoke hook.

        This is similar to :meth:`.Command.before_invoke`.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass

    @_cog_special_method
    async def cog_after_invoke(self, ctx: Context) -> None:
        """A special method that acts as a cog local post-invoke hook.

        This is similar to :meth:`.Command.after_invoke`.

        This **must** be a coroutine.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context.
        """
        pass



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


import re

from typing import Any, Optional
from fortnitepy.user import User
from fortnitepy.party import PartyMember
from fortnitepy.friend import Friend

from .errors import BadArgument
from .context import Context


__all__ = (
    'Converter',
    'IDConverter',
    'UserConverter',
    'PartyMemberConverter',
    'FriendConverter',
    'IDConverter',
    'Greedy',
)


class Converter:
    """The base class of custom converters that require the :class:`.Context`
    to be passed to be useful.

    This allows you to implement converters that function similar to the
    special cased ``fortnitepy`` classes.

    Classes that derive from this should override the
    :meth:`~.Converter.convert` method to do its conversion logic. This method
    must be a coroutine.
    """

    async def convert(self, ctx: 'Context', argument: str) -> Any:
        """|coro|

        The method to override to do conversion logic.

        If an error is found while converting, it is recommended to
        raise a :exc:`.CommandError` derived exception as it will
        properly propagate to the error handlers.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context that the argument is being used in.
        argument: :class:`str`
            The argument that is being converted.
        """
        raise NotImplementedError('Derived classes need to implement this.')


class IDConverter(Converter):
    def __init__(self) -> None:
        self._id_regex = re.compile(r'([0-9a-fA-F]{32})$')

    def _get_id_match(self, argument: str) -> Any:
        return self._id_regex.match(argument)


class UserConverter(IDConverter):
    """Converts to a :class:`~fortnitepy.User`.

    The lookup strategy is as follows (in order):
    1. Cache lookup by ID.
    2. Cache lookup by display name.
    3. API Request to fetch the user by id/display name.
    """

    async def convert(self, ctx: Context, argument: str) -> User:
        bot = ctx.bot
        match = self._get_id_match(argument)

        if match is not None:
            to_lookup = match.group(1).lower()
        else:
            to_lookup = argument

        result = await bot.fetch_user(to_lookup, cache=True)

        if result is None:
            raise BadArgument('User "{}" not found'.format(argument))

        return result


class PartyMemberConverter(IDConverter):
    """Converts to a :class:`~fortnitepy.PartyMember`.

    All lookups are done via the bots party member cache.

    The lookup strategy is as follows (in order):
    1. Lookup by ID.
    2. Lookup by display name.
    """

    async def convert(self, ctx: Context, argument: str) -> PartyMember:
        bot = ctx.bot
        match = self._get_id_match(argument)
        party = ctx.party or bot.party

        result = None
        if match is None:
            for member in party._members.values():
                if member.display_name.casefold() == argument.casefold():
                    result = member
                    break
        else:
            user_id = match.group(1).lower()
            result = party.get_member(user_id)

        if result is None:
            raise BadArgument('Party member "{}" not found'.format(argument))

        return result


class FriendConverter(IDConverter):
    """Converts to a :class:`~fortnitepy.Friend`.

    All lookups are via the friend cache.

    The lookup strategy is as follows (in order):
    1. Lookup by ID.
    2. Lookup by display name.
    """

    async def convert(self, ctx: Context, argument: str) -> Friend:
        bot = ctx.bot
        match = self._get_id_match(argument)

        result = None
        if match is not None:
            user_id = match.group(1).lower()
            result = bot.get_friend(user_id)
        else:
            for friend in bot._friends.values():
                if friend.display_name.casefold() == argument.casefold():
                    result = friend
                    break

        if result is None:
            raise BadArgument('Friend "{}" not found'.format(argument))

        return result


class _Greedy:
    __slots__ = ('converter',)

    def __init__(self, *, converter: Optional[Converter] = None) -> None:
        self.converter = converter

    def __getitem__(self, params: Any) -> Any:
        if not isinstance(params, tuple):
            params = (params,)
        if len(params) != 1:
            raise TypeError('Greedy[...] only takes a single argument')

        converter = params[0]

        if not (callable(converter) or isinstance(converter, Converter)
                or hasattr(converter, '__origin__')):
            raise TypeError('Greedy[...] expects a type or a Converter '
                            'instance.')

        NoneType = type(None)
        if converter is str or converter is NoneType or converter is _Greedy:
            raise TypeError('Greedy[%s] is invalid.' % converter.__name__)

        return self.__class__(converter=converter)


Greedy = _Greedy()



# -*- coding: utf-8 -*-

"""
MIT License

Copyright (c) 2019-2021 Terbau

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime

from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from .client import Client
    from .friend import Friend
    from .party import ClientParty, PartyMember


class MessageBase:

    __slots__ = ('_client', '_author', '_content', '_created_at')

    def __init__(self, client: 'Client',
                 author: Union['Friend', 'PartyMember'],
                 content: str) -> None:
        self._client = client
        self._author = author
        self._content = content
        self._created_at = datetime.datetime.utcnow()

    @property
    def client(self) -> 'Client':
        """:class:`Client`: The client."""
        return self._client

    @property
    def author(self) -> 'Friend':
        """:class:`Friend`: The author of the message."""
        return self._author

    @property
    def content(self) -> str:
        """:class:`str`: The content of the message."""
        return self._content

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: The time of when this message was
        received in UTC.
        """
        return self._created_at


class FriendMessage(MessageBase):

    __slots__ = MessageBase.__slots__

    def __init__(self, client: 'Client',
                 author: 'Friend',
                 content: str) -> None:
        super().__init__(client, author, content)

    def __repr__(self) -> str:
        return ('<FriendMessage author={0.author!r} '
                'created_at={0.created_at!r}>'.format(self))

    async def reply(self, content: str) -> None:
        """|coro|

        Replies to the message with the given content.

        Parameters
        ----------
        content: :class:`str`
            The content of the message
        """
        return await self.author.send(content)


class PartyMessage(MessageBase):

    __slots__ = MessageBase.__slots__ + ('party',)

    def __init__(self, client: 'Client',
                 party: 'ClientParty',
                 author: 'PartyMember',
                 content: str) -> None:
        super().__init__(client, author, content)
        self.party = party

    def __repr__(self) -> str:
        return ('<PartyMessage party={0.party!r} author={0.author!r} '
                'created_at={0.created_at!r}>'.format(self))

    @property
    def author(self) -> 'PartyMember':
        """:class:`PartyMember`: The author of a message."""
        return self._author

    async def reply(self, content: str) -> None:
        """|coro|

        Replies to the message with the given content.

        Parameters
        ----------
        content: :class:`str`
            The content of the message
        """
        return await self.party.send(content)
